# Copyright (c) 2017, University of Bristol and University of Southampton.

from EStat.Templating import Template

class TemplateSAEModel(Template):
    'Uses the 2LevelSAEMCMC template to fit SAE models using MCMC and constructs lots of related graphs for use in the SAE eBook.'

    __version__ = '1.0.0'

    tags = [ 'Data manipulation', 'SAE', 'Model' ]
    engines = ['Python_script']

    inputs = '''
popdata = DataFileChooser('Population dataset: ')
resp = DataVector('Response variable: ')
D = Text('Specify distribution: ', ['Normal', 'Binomial', 'Poisson', '-ve Binomial'])
if D == 'Binomial':
    n = DataVector('Denominator: ',  help='e.g. if modelling a binary 0/1 response, select a constant of ones.')
    link = Text('Specify link function: ', ['logit', 'probit', 'cloglog'])
if D == 'Poisson':
    link = Text(value = 'ln')
    offset = Boolean('Is there an offset: ', help="An offset allows you to model rates, instead of raw counts.")
    if offset:
        n = DataVector('Offset: ')
if D == '-ve Binomial':
    offset = Boolean('Is there an offset: ')
    if offset:
        n = DataVector('Offset: ')
if D == 'Normal':
    transform = Text('Transformation:', ['Box-Cox', 'Log', 'None'])
    if transform == 'Box-Cox':
        lambtype = Text('Lambda:', ['Automatic', 'Modelled', 'Manual'])
        if lambtype == 'Manual':
            lambval = Text('Lambda parameter for response tranformation: ')
else:
    transform = Text(value='None')

idcol = IDVector('Common ID variable: ')
pred = DataMatrix('Common predictor variables', allow_cat = True)

povind = Boolean('Do you want to calculate poverty related estimates, e.g. Head count ratio?: ')
if povind:
    userthresh = Text('Threshold value: ', ['Automatic', 'Manual'])
    if userthresh == 'Manual':
        thresh = Text('Threshold value')
    else:
        thresh = Text(value='')
ineqind = Boolean('Do you want to calculate inequality related estimates, e.g. GINI index?: ')

seed = Text('Random Seed: ', default=1)
nchain = Integer('Number of chains: ', min=1, default=3)
burnin = Integer('Length of burnin: ', min=0, default=500)
niter = Integer('Number of iterations: ', min=0, default=2000)
thin = Integer('Thinning: ', min=1, default=1)
nproc = Integer('Parallel cores for predictions: ', min=1, default=6)
# Note need to ensure pred is filled in before poppred line
z = list(pred)
poppred = DataMatrix(value=list(pred.orignames), categorical=pred.categorical, data_src=popdata, allow_cat = True)
'''


    pythonscript = '''
import numpy
import scipy
import scipy.stats

from io import BytesIO

from matplotlib.figure import Figure
import matplotlib.lines as lines
import matplotlib.patches as patches
from matplotlib.backends.backend_agg import FigureCanvasAgg

import sys
import os
sys.path.append(os.path.abspath(os.path.join('..', '..', 'apps', 'cmdtest')))
from cmdtest import RunStatJR

import EStat
from EStat.Templating import *
from EStat.DTAFile import DTAFile

data1 = datafile.copy()
data1.addvariable("_intercept", data=numpy.ones(datafile.nobs))

data2 = context[str(popdata)].copy()
data2.addvariable("_intercept", data=numpy.ones(data2.nobs))

datasets.datasets['workdata'] = data1
datasets.datasets['workdata2'] = data2

sampvars = []
for var in pred:
    if var in datafile.variables.keys():
        sampvars.append(datafile.variables[var]['data'])
    else:
        # Variable names do not match
        print("Variable " + var + " not found in sample dataset")
sampvars.append(datafile.variables[str(idcol)]['data'])
sampvars.append(numpy.ones(datafile.nobs))

popvars = []
for var in pred:
    if var in data2.variables.keys():
        popvars.append(data2.variables[var]['data'])
    else:
        # Variable names do not match
        print("Variable " + var + " not found in population dataset")
popvars.append(data2.variables[str(idcol)]['data'])
popvars.append(numpy.zeros(data2.nobs))

print("Sample:")
print(numpy.array(sampvars))

print("Population")
print(numpy.array(popvars))

allvars = numpy.concatenate((sampvars, popvars), axis=1)

print("Joined:")
print(allvars)

retval = DTAFile()
retval.nobs = allvars.shape[1]
for i in range(0, len(pred)):
    retval.addvariable(pred[i], data = allvars[i])
    retval.variables[pred[i]]['varlabel'] = datafile.variables[pred[i]]['varlabel']
    lab = datafile.variables[pred[i]]['label']
    retval.variables[pred[i]]['label'] = lab
    if lab in datafile.valuelabels.keys():
        retval.valuelabels[lab] = datafile.valuelabels[lab]

retval.addvariable(str(idcol), data = numpy.array(allvars[-2], dtype=int))
retval.variables[str(idcol)]['varlabel'] = datafile.variables[str(idcol)]['varlabel']
lab = datafile.variables[str(idcol)]['label']
retval.variables[str(idcol)]['label'] = lab
if lab in datafile.valuelabels.keys():
    retval.valuelabels[lab] = datafile.valuelabels[lab]
if lab in data2.valuelabels.keys():
    retval.valuelabels[lab].update(data2.valuelabels[lab])

retval.addvariable('_sample_ind', data = numpy.array(allvars[-1], dtype=int))

outputs['result'] = retval

uniqids = numpy.unique(retval.variables[str(idcol)]['data'])

uniqpopids = numpy.unique(data2.variables[str(idcol)]['data'])
uniqsampids = numpy.unique(datafile.variables[str(idcol)]['data'])

pred.name.insert(0, '_intercept')

inputs = {'y': str(resp), 'D': str(D), 'L2ID': str(idcol), 'x': ','.join(pred), 'popdata': 'workdata2'}
inputs.update({'numproc': str(nproc), 'transform': str(transform), 'povind': str(povind), 'ineqind': str(ineqind)})
if povind:
    inputs['userthresh'] = str(userthresh)
    if userthresh:
        inputs['thresh'] = str(thresh)

if D == 'Normal':
    if transform == 'Box-Cox':
        inputs['lambtype'] = str(lambtype)
        if lambtype == 'Manual':
            inputs['lambval'] = str(lambval)
if D == 'Binomial':
    inputs['n'] = str(n)
    inputs['link'] = str(link)
if D == 'Poisson':
    inputs['offset'] = str(offset)
    if offset:
        inputs['n'] = str(n)
if D == '-ve Binomial':
    inputs['offset'] = str(offset)
    if offset:
        inputs['n'] = str(n)
estinputs = {'iterations': str(niter), 'burnin': str(burnin), 'defaultsv': 'Yes', 'thinning': str(thin), 'nchains': str(nchain)}
estinputs.update({'defaultalg': 'Yes', 'outdata': 'modelchains', 'seed': str(seed), 'makepred': 'Yes'})
m = RunStatJR(template='2LevelSAEMCMC', dataset='workdata', invars = inputs, estoptions = estinputs)

outputs['equation.tex'] = m.eng.outputs['equation.tex']

# SAE estimates
saeests = m.eng.outputs[str(idcol)]
modelchains = m.eng.outputs['modelchains']

class ParameterValue:
    def __init__(self, m):
        self.dta = m.eng.outputs['modelparameters.dta']
        self.parammap = {key:value for (key, value) in zip(self.dta.variables['parameter']['data'], range(0, self.dta.nobs))}

    def __call__(self, *args, **kwargs):
        param = args[0]
        type = args[1]
        return self.dta.variables[type]['data'][self.parammap[param]]

getresult = ParameterValue(m)

tabout = TabularOutput()
tabout.column_headings = ['Parameter', 'Variable', 'Mean', 'SD', 'ESS']
tabout.add_row('beta_0', ['Intercept', getresult('beta_0' , 'mean'), getresult('beta_0', 'sd'), getresult('beta_0' , 'ESS')])
for i in range(1, len(pred)):
    tabout.add_row('beta_' + str(i), [pred[i], getresult('beta_' + str(i), 'mean'), getresult('beta_' + str(i), 'sd'), getresult('beta_' + str(i), 'ESS')])

if D == 'Normal':
    tabout.add_row('sigma2', ['Level-1 variance', getresult('sigma2', 'mean'), getresult('sigma2', 'sd'), getresult('sigma2', 'ESS')])

tabout.add_row('sigma_u', ['Level-2 variance', getresult('sigma_u', 'mean'), getresult('sigma_u', 'sd'), getresult('sigma_u', 'ESS')])

if transform == 'Box-Cox':
    if lambtype == 'Modelled':
        tabout.add_row('lamb', ['Box-Cox Lambda', getresult('lamb', 'mean'), getresult('lamb', 'sd'), getresult('lamb', 'ESS')])

outputs['sae_modelresults'] = tabout

signiflist = []
for i in range(1, len(pred)):
    if (getresult('beta_' + str(i), 'sd') * 1.96) < numpy.abs(getresult('beta_' + str(i), 'mean')):
        signiflist.append(pred[i])

if len(signiflist) > 1:
    signiflist[-1] = ' and ' + signiflist[-1]

html = '<p>'
if len(signiflist) == 0:
    html += 'Here we see that no predictors have a significant effect on the response variable. '
if len(signiflist) == 1:
    html += 'Here we see that in the sample the predictor ' + ', '.join(signiflist) + ' has a significant effect on the response variable. '
if len(signiflist) > 1:
    html += 'Here we see that in the sample the predictors ' + ', '.join(signiflist) + ' have significant effects on the response variable. '
html += '</p>'
outputs['modeltext.html'] = HTMLOutput(html, "Text describing model")

# Residuals
uniqresi = numpy.unique(datafile.variables[str(idcol)]['data'])

#residuals = numpy.zeros((nchain*niter, len(uniqresi)))
resmean = numpy.zeros(len(uniqresi))
#ressd = numpy.zeros(len(uniqresi))

for i in range(0, len(uniqresi)):
    resmean[i] = getresult('u_' + str(i), 'mean')
    #residuals[:, i] = modelchains.variables['u_' + str(i)]['data']
    #resmean[i] = modelchains.variables['u_' + str(i)]['data'].mean()
    #ressd[i] = modelchains.variables['u_' + str(i)]['data'].std()

fig = Figure(figsize=(8,8))
ax = fig.add_subplot(100 + 10 + 1, xlabel = 'Theoretical Quantiles', ylabel = 'Residuals')
scipy.stats.probplot(resmean, plot=ax)
canvas = FigureCanvasAgg(fig)
buf = BytesIO()
canvas.print_figure(buf, dpi=80, format='svg')
buf.seek(0)
outputs['residuals.svg'] = ImageOutput(buf.getvalue())
buf.close()

html = '<p>'
html += 'Although the estimates are interesting our primary interest is in the small area estimates that this model produces. '
html += 'The table below gives estimates for the mean and SD for each of the ' + str(len(uniqids)) + ' small areas in the dataset. '
html += 'These estimates can be found in the columns headed population whilst for comparison in the columns headed sample are the means and SDS for ' + str(resp) + ' just using the sample data. '
html += '</p>'
outputs['prediction_text.html'] = HTMLOutput(html, "Text describing prediction table")

tabout = TabularOutput()
tabout.column_headings = ['Code', 'Name', 'Sample mean', 'Sample sd', 'Population mean', 'Population sd']

#predout = outputs['prediction_datafile']
#predout2 = outputs['prediction_' + str(popdata)]

fig = Figure(figsize=(8,24))
ax = fig.add_subplot(100 + 10 + 1)
pltrow = []

def num2str(val):
    import numpy
    if numpy.isnan(val):
        return '-'
    else:
        return str(val)

ymean = numpy.zeros(data2.nobs)
for i in range(0, data2.nobs):
    ymean[i] = getresult('y_pred_' + str(i), 'mean')

for id in uniqids:
    if lab in retval.valuelabels and id in retval.valuelabels[lab]:
        name = retval.valuelabels[lab][id]
    else:
        name = str(idcol) + '==' + str(id)
    samppred = datafile.variables[str(resp)]['data'][datafile.variables[str(idcol)]['data'] == id]
    sampmean = num2str(samppred.mean())
    sampsd = num2str(samppred.std())
    poppred = ymean[data2.variables[str(idcol)]['data'] == id]
    #popmean = num2str(poppred.mean())
    u_id = numpy.nonzero(uniqids == id)[0][0]
    popmean = num2str(getresult('yj_mean_' + str(u_id), 'mean'))
    popsd = num2str(getresult('yj_std_' + str(u_id), 'mean'))

    tabout.add_row(str(id), [name, sampmean, sampsd, popmean, popsd])

    pltrow.append(samppred)
    pltrow.append(poppred)

outputs['prediction_summary'] = tabout

bplot = ax.boxplot(pltrow, vert=False, patch_artist=True)
for patch, colour in zip(bplot['boxes'], ['pink', 'green']*len(uniqids)):
    patch.set_facecolor(colour)

ax.set_xlabel(str(resp))
yvals = numpy.arange(0, len(uniqids) * 2, 2) + 1.5
ax.set_yticks(yvals)
if lab in retval.valuelabels and id in uniqids:
    ylabs = [retval.valuelabels[lab][id] for id in uniqids]
else:
    ylabs = [str(idcol) + '==' + str(id) for id in uniqids]
ax.set_yticklabels(ylabs)

ax.legend(handles=[patches.Patch(color='pink', label='Sample'), patches.Patch(color='green', label='Population')], bbox_to_anchor=(0., 1.02, 1., .102), loc=3, ncol=2, mode="expand", borderaxespad=0.)

canvas = FigureCanvasAgg(fig)
fig.tight_layout()

buf = BytesIO()
canvas.print_figure(buf, dpi=80, format='svg')

buf.seek(0)
outputs['box_vert.svg'] = ImageOutput(buf.getvalue())

fig = Figure(figsize = (16, 8))
ax = fig.add_subplot(100 + 20 + 1, title = 'Prediction for ' + str(resp) + ' (quantiles)', xlabel = str(idcol), ylabel = str(resp))
ax.plot(saeests.variables[str(idcol)]['data'], saeests.variables['yj_q10']['data'], label = '10% quantile')
ax.plot(saeests.variables[str(idcol)]['data'], saeests.variables['yj_q25']['data'], label = '25% quantile')
ax.plot(saeests.variables[str(idcol)]['data'], saeests.variables['yj_q50']['data'], label = '50% quantile')
ax.plot(saeests.variables[str(idcol)]['data'], saeests.variables['yj_q75']['data'], label = '75% quantile')
ax.plot(saeests.variables[str(idcol)]['data'], saeests.variables['yj_q90']['data'], label = '90% quantile')
ax.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)

canvas = FigureCanvasAgg(fig)
buf = BytesIO()
canvas.print_figure(buf, dpi=80, format='svg')

buf.seek(0)
outputs['sae_quantiles.svg'] = ImageOutput(buf.getvalue())
buf.close()

tabout = TabularOutput()
tabout.column_headings = ['Code', 'Name', 'Q10', 'Q25', 'Q50', 'Q75', 'Q90']

for id in uniqids:
    if lab in retval.valuelabels and id in retval.valuelabels[lab]:
        name = retval.valuelabels[lab][id]
    else:
        name = str(idcol) + '==' + str(id)

    q10 = saeests.variables['yj_q10']['data'][saeests.variables[str(idcol)]['data'] == id][0]
    q25 = saeests.variables['yj_q25']['data'][saeests.variables[str(idcol)]['data'] == id][0]
    q50 = saeests.variables['yj_q50']['data'][saeests.variables[str(idcol)]['data'] == id][0]
    q75 = saeests.variables['yj_q75']['data'][saeests.variables[str(idcol)]['data'] == id][0]
    q90 = saeests.variables['yj_q90']['data'][saeests.variables[str(idcol)]['data'] == id][0]
    tabout.add_row(str(id), [name, q10, q25, q50, q75, q90])

outputs['sae_quantilestable'] = tabout


fig = Figure(figsize = (8, 8))
ax = fig.add_subplot(100 + 10 + 1, title = 'Prediction for ' + str(resp) + ' (mean)', xlabel = str(idcol), ylabel = str(resp))
ax.errorbar(saeests.variables[str(idcol)]['data'], saeests.variables['yj_mean']['data'], yerr = 1.96 * saeests.variables['yj_mean_sd']['data'], fmt = 'x')

canvas = FigureCanvasAgg(fig)
buf = BytesIO()
canvas.print_figure(buf, dpi=80, format='svg')

buf.seek(0)
outputs['sae_mean.svg'] = ImageOutput(buf.getvalue())
buf.close()

if povind:
    if userthresh == 'Automatic':
        thresh = str(0.6 * numpy.median(data1.variables[resp]['data'])) # Recalculated, so need to keep up to date if underlying template changes.
    html = '<p>'
    html += 'Next we have the Head count ratio (HCR). '
    html += 'This is again often associated with income and poverty and represents the proportion of each small area whose value of ' + str(resp) + ' lies below a pre-specified threshold, in this case ' + str(thresh) + '.'
    html += '</p>'
    outputs['sae_headcountdesc.html'] = HTMLOutput(html, "Head count description")

    html = '<p>'
    html += 'Below is a graph of the head count ratio (with credible intervals) for the ' + str(len(uniqids)) + ' small areas.'
    html += '</p>'
    outputs['sae_headcounttext.html'] = HTMLOutput(html, "Head count text")

    fig = Figure(figsize = (8, 8))
    ax = fig.add_subplot(100 + 10 + 1, title = 'Prediction for ' + str(resp) + ' (head count)', xlabel = str(idcol), ylabel = 'Head Count')
    ax.errorbar(saeests.variables[str(idcol)]['data'], saeests.variables['yj_headcount']['data'], yerr = 1.96 * saeests.variables['yj_headcount_sd']['data'], fmt = 'x')

    canvas = FigureCanvasAgg(fig)
    buf = BytesIO()
    canvas.print_figure(buf, dpi=80, format='svg')

    buf.seek(0)
    outputs['sae_headcount.svg'] = ImageOutput(buf.getvalue())
    buf.close()

    html = '<p>'
    html += 'Below is a graph of the poverty gap index (with credible intervals) for the ' + str(len(uniqids)) + ' small areas.'
    html += '</p>'
    outputs['sae_povgaptext.html'] = HTMLOutput(html, "Poverty gap text")

    fig = Figure(figsize = (8, 8))
    ax = fig.add_subplot(100 + 10 + 1, title = 'Prediction for ' + str(resp) + ' (poverty gap)', xlabel = str(idcol), ylabel = 'Poverty Gap')
    ax.errorbar(saeests.variables[str(idcol)]['data'], saeests.variables['yj_povgap']['data'], yerr = 1.96 * saeests.variables['yj_povgap_sd']['data'], fmt = 'x')

    canvas = FigureCanvasAgg(fig)
    buf = BytesIO()
    canvas.print_figure(buf, dpi=80, format='svg')

    buf.seek(0)
    outputs['sae_povgap.svg'] = ImageOutput(buf.getvalue())
    buf.close()

if ineqind:
    html = '<p>'
    html += 'Below is a graph of the Gini index (with credible intervals) for the ' + str(len(uniqids)) + ' small areas.'
    html += '</p>'
    outputs['sae_ginitext.html'] = HTMLOutput(html, "GINI text")

    fig = Figure(figsize = (8, 8))
    ax = fig.add_subplot(100 + 10 + 1, title = 'Prediction for ' + str(resp) + ' (GINI)', xlabel = str(idcol), ylabel = 'GINI')
    ax.errorbar(saeests.variables[str(idcol)]['data'], saeests.variables['yj_gini']['data'], yerr = 1.96 * saeests.variables['yj_gini_sd']['data'], fmt = 'x')

    canvas = FigureCanvasAgg(fig)
    buf = BytesIO()
    canvas.print_figure(buf, dpi=80, format='svg')

    buf.seek(0)
    outputs['sae_gini.svg'] = ImageOutput(buf.getvalue())
    buf.close()

    html = '<p>'
    html += 'The quintile share ratio (QSR) measures the ratio of the total of the response (usually income) received by the top 20% of the population to the total of the response received by the bottom 20% of the population. '
    html += 'So a value of 2 would mean that on average individuals in the top 20% would have values of ' + str(resp) + ' twice as high as individuals in the bottom 20%. '
    html += '</p>'
    outputs['sae_qsharedesc.html'] = HTMLOutput(html, "Quintile share description")

    html = '<p>'
    html += 'Below is a graph of the quintile share ratio (with credible intervals) for the ' + str(len(uniqids)) + ' small areas. '
    html += '</p>'
    outputs['sae_qsharetext.html'] = HTMLOutput(html, "Quintile share text")

    fig = Figure(figsize = (8, 8))
    ax = fig.add_subplot(100 + 10 + 1, title = 'Prediction for ' + str(resp) + ' (quintile share)', xlabel = str(idcol), ylabel = 'Quintile Share')
    ax.errorbar(saeests.variables[str(idcol)]['data'], saeests.variables['yj_qshare']['data'], yerr = saeests.variables['yj_qshare_sd']['data'], fmt = 'x')

    canvas = FigureCanvasAgg(fig)
    buf = BytesIO()
    canvas.print_figure(buf, dpi=80, format='svg')

    buf.seek(0)
    outputs['sae_qshare.svg'] = ImageOutput(buf.getvalue())
    buf.close()

tabout = TabularOutput()
tabout.column_headings = ['Code', 'Area']
if ineqind:
    tabout.column_headings += ['Gini', 'Gini rank', 'QSR', 'QSR rank']

if povind:
    tabout.column_headings += ['HCR', 'HCR rank', 'PGI', 'PGI rank']

if ineqind:
    giniranks = scipy.stats.rankdata(saeests.variables['yj_gini']['data'])
    qsrranks = scipy.stats.rankdata(saeests.variables['yj_qshare']['data'])

if povind:
    hcrranks = scipy.stats.rankdata(saeests.variables['yj_headcount']['data'])
    pgiranks = scipy.stats.rankdata(saeests.variables['yj_povgap']['data'])

if povind or ineqind:
    for id in uniqids:
        if lab in retval.valuelabels and id in retval.valuelabels[lab]:
            name = retval.valuelabels[lab][id]
        else:
            name = str(idcol) + '==' + str(id)
        row = [name]
        if ineqind:
            gini = saeests.variables['yj_gini']['data'][saeests.variables[str(idcol)]['data'] == id][0]
            ginirank = giniranks[saeests.variables[str(idcol)]['data'] == id][0]
            qsr = saeests.variables['yj_qshare']['data'][saeests.variables[str(idcol)]['data'] == id][0]
            qsrrank = qsrranks[saeests.variables[str(idcol)]['data'] == id][0]
            row += [gini, ginirank, qsr, qsrrank]
        if povind:
            hcr = saeests.variables['yj_headcount']['data'][saeests.variables[str(idcol)]['data'] == id][0]
            hcrrank = hcrranks[saeests.variables[str(idcol)]['data'] == id][0]
            pgi = saeests.variables['yj_povgap']['data'][saeests.variables[str(idcol)]['data'] == id][0]
            pgirank = pgiranks[saeests.variables[str(idcol)]['data'] == id][0]
            row += [hcr, hcrrank, pgi, pgirank]

        tabout.add_row(str(id), row)

    outputs['sae_summarystats'] = tabout

#print("Copying outputs")
#for rsc in m.eng.inputs.keys():
#    outputs[rsc] = m.eng.inputs[rsc]

#for rsc in m.eng.outputs.keys():
#    outputs[rsc] = m.eng.outputs[rsc]

# outputs.update(m.eng.outputs)

#print("Copying outputs done")
'''