# Copyright (c) 2017, University of Bristol and University of Southampton.

from EStat.Templating import Template

class TemplateSAEPredictor(Template):
    'Explores predictors variables for SAE models and is used in the SAE eBook.'

    __version__ = '1.0.0'

    tags = [ 'Data manipulation', 'SAE' ]
    engines = ['Python_script']

    inputs = '''
popdata = DataFileChooser('Population dataset: ')
idcol = IDVector('Common ID variable: ')
pred = DataVector('Predictor variable')
'''


    pythonscript = '''
import numpy

from io import BytesIO

from matplotlib.figure import Figure
import matplotlib.lines as lines
import matplotlib.patches as patches
from matplotlib.backends.backend_agg import FigureCanvasAgg

import sys
import os
sys.path.append(os.path.abspath(os.path.join('..', '..', 'apps', 'cmdtest')))
from cmdtest import RunStatJR

import EStat
from EStat.Templating import *
from EStat.DTAFile import DTAFile

data2 = globals()[str(popdata)]

datafile.addvariable("Intercept", data=numpy.ones(datafile.nobs))
data2.addvariable("Intercept", data=numpy.ones(data2.nobs))

datasets.datasets['workdata'] = datafile
datasets.datasets['workdata2'] = data2

sampvars = []
if pred in datafile.variables.keys():
    sampvars.append(datafile.variables[pred]['data'])
else:
    # Variable names do not match
        print("Variable do not match between datasets")
sampvars.append(datafile.variables[str(idcol)]['data'])
sampvars.append(numpy.ones(datafile.nobs))

popvars = []
if pred in data2.variables.keys():
    popvars.append(data2.variables[pred]['data'])
else:
    # Variable names do not match
        print("Variable do not match between datasets")
popvars.append(data2.variables[str(idcol)]['data'])
popvars.append(numpy.zeros(data2.nobs))

print("Sample:")
print(numpy.array(sampvars))

print("Population")
print(numpy.array(popvars))

allvars = numpy.concatenate((sampvars, popvars), axis=1)

print("Joined:")
print(allvars)

retval = DTAFile()
retval.nobs = allvars.shape[1]
retval.addvariable(pred, data = allvars[0])
retval.variables[pred]['varlabel'] = datafile.variables[pred]['varlabel']
lab = datafile.variables[pred]['label']
retval.variables[pred]['label'] = lab
if lab in datafile.valuelabels.keys():
    retval.valuelabels[lab] = datafile.valuelabels[lab]

retval.addvariable(str(idcol), data = numpy.array(allvars[-2], dtype=int))
retval.variables[str(idcol)]['varlabel'] = datafile.variables[str(idcol)]['varlabel']
lab = datafile.variables[str(idcol)]['label']
retval.variables[str(idcol)]['label'] = lab
if lab in datafile.valuelabels.keys():
    retval.valuelabels[lab] = datafile.valuelabels[lab]
if lab in data2.valuelabels.keys():
    retval.valuelabels[lab].update(data2.valuelabels[lab])

retval.addvariable('_sample_ind', data = numpy.array(allvars[-1], dtype=int))

outputs['result'] = retval

uniqids = numpy.unique(retval.variables[str(idcol)]['data'])

uniqpopids = numpy.unique(data2.variables[str(idcol)]['data'])
uniqsampids = numpy.unique(datafile.variables[str(idcol)]['data'])

html = '<p>'
html += 'We can start by looking at the distribution of ' + str(pred) + ' in the sample dataset (as a whole) and the population. '
html += 'In the plots below we can see to the left the population and to the right the sample data with superimposed a best fitting normal curve. '
html += 'Although the normality assumption in our modelling is not related to the predictors we might still want to transform them if they are skewed as in this case outlying data points might have significant influence on the relationship used in the SAE modelling. '
html += 'Note also that if the predictor variable is categorical this plot may be less informative '
html += 'Here aside from some areas having no sample data we can look and see how representative the samples in other areas are of the population. '
html += '</p>'

outputs['histogramtext.html'] = HTMLOutput(html, "Summary text for predictor")

sampvars = retval.variables[pred]['data'][retval.variables['_sample_ind']['data'] == 1]
sampweights = numpy.ones(len(sampvars)) / float(len(sampvars)) * 100.0
sampprop = sampvars * sampweights
popvars = retval.variables[pred]['data'][retval.variables['_sample_ind']['data'] == 0]
popweights = numpy.ones(len(popvars)) / float(len(popvars)) * 100.0
popprop = popvars * popweights

popnormx = numpy.linspace(popvars.min(), popvars.max(), 100)
popnormy = scipy.stats.norm.pdf(popnormx, popvars.mean(), popvars.std()) * len(popvars) * (popvars.max() - popvars.min()) / 20.0

sampnormx = numpy.linspace(sampvars.min(), sampvars.max(), 100)
sampnormy = scipy.stats.norm.pdf(sampnormx, sampvars.mean(), sampvars.std()) * len(sampvars) * (sampvars.max() - sampvars.min()) / 20.0

fig = Figure(figsize=(8,4))
ax = fig.add_subplot(1, 2, 1)
ax.set_title('Population')
ax.set_xlabel(pred, fontsize=7)
ax.set_ylabel('Frequency', fontsize=7)
for label in (ax.get_xticklabels() + ax.get_yticklabels()):
    label.set_fontsize(5)
ax.hist(popvars, 20)
ax.plot(popnormx, popnormy, linewidth=3)


ax = fig.add_subplot(1, 2, 2)
ax.set_title('Sample')
ax.set_xlabel(pred, fontsize=7)
ax.set_ylabel('Frequency', fontsize=7)
for label in (ax.get_xticklabels() + ax.get_yticklabels()):
    label.set_fontsize(5)
ax.hist(sampvars, 20)
ax.plot(sampnormx, sampnormy, linewidth=3)

canvas = FigureCanvasAgg(fig)
fig.tight_layout()

buf = BytesIO()
canvas.print_figure(buf, dpi=80, format='svg')

buf.seek(0)
outputs['histogram.svg'] = ImageOutput(buf.getvalue())
buf.close()


fig = Figure(figsize=(4,4))
ax = fig.add_subplot(100 + 10 + 1)
ax.set_xlabel(pred, fontsize=7)
ax.set_ylabel('Percentage', fontsize=7)
for label in (ax.get_xticklabels() + ax.get_yticklabels()):
    label.set_fontsize(5)
ax.hist(popvars, 20, weights = popweights, histtype='step', fc='none', label='Population')
ax.hist(sampvars, 20, weights = sampweights, histtype='step', fc='none', label='Sample')
#ax.plot(popnormx, popnormy, linewidth=3)
#ax.plot(sampnormx, sampnormy, linewidth=3)
box = ax.get_position()
ax.set_position([box.x0, box.y0, box.width * 0.2, box.height])
ax.legend(loc = 'upper center', mode = "expand", ncol = 2, bbox_to_anchor = (0, 1.02, 1, 0.2), frameon = False)

canvas = FigureCanvasAgg(fig)
fig.tight_layout()

buf = BytesIO()
canvas.print_figure(buf, dpi=80, format='svg')

buf.seek(0)
outputs['histogram_alt.svg'] = ImageOutput(buf.getvalue())
buf.close()

fig = Figure(figsize=(8,8))

nrows = int(numpy.ceil(numpy.sqrt(len(uniqids))))
ncols = int(numpy.ceil(float(len(uniqids)) / float(nrows)))

print("nrows: " + str(nrows) + " ncols: " + str(ncols))

#for i in range(0, len(uniqids)):
#    id = uniqids[i]
#    samp = retval.variables[pred]['data'][(retval.variables['_sample_ind']['data'] == 1) & (retval.variables[str(idcol)]['data'] == id)]
#    pop = retval.variables[pred]['data'][(retval.variables['_sample_ind']['data'] == 0) & (retval.variables[str(idcol)]['data'] == id)]
#    ax = fig.add_subplot(nrows, ncols, i+1)
#    ax.boxplot([samp, pop], labels = ['Sample', 'Population'])

#axs = fig.subplots(nrows = nrows, ncols = ncols, sharex = True, sharey = True)

#idx = 0
#for row in range(0, nrows):
#    for col in range(0, ncols):
#        ax = axs[row, col]
#        ax.set_xticklabels(['Sample', 'Population'], rotation='vertical')
#
#        if idx >= len(uniqids):
#            continue
#        id = uniqids[idx]
#        #print("row: " + str(row) + " col: " + str(col) + " idx: " + str(idx) + " id: " + str(id))
#        samp = retval.variables[pred]['data'][(retval.variables['_sample_ind']['data'] == 1) & (retval.variables[str(idcol)]['data'] == id)]
#        pop = retval.variables[pred]['data'][(retval.variables['_sample_ind']['data'] == 0) & (retval.variables[str(idcol)]['data'] == id)]
#        ax.boxplot([samp, pop]) # , labels = ['Sample', 'Population'])
#        idx += 1

#canvas = FigureCanvasAgg(fig)
#fig.tight_layout()

#buf = BytesIO()
#canvas.print_figure(buf, dpi=80, format='svg')

#buf.seek(0)
#outputs['box.svg'] = ImageOutput(buf.getvalue())
#buf.close()




#fig = Figure(figsize=(8,8))
#axs = fig.subplots(nrows = nrows, ncols = 1, sharex = True, sharey = True)
#
#idx = 0
#for row in range(0, nrows):
#    pltrow = []
#    for col in range(0, ncols):
#        if idx >= len(uniqids):
#                continue
#        id = uniqids[idx]
#        #print("row: " + str(row) + " col: " + str(col) + " idx: " + str(idx) + " id: " + str(id))
#        pltrow.append(retval.variables[pred]['data'][(retval.variables['_sample_ind']['data'] == 1) & (retval.variables[str(idcol)]['data'] == id)]) # samp
#        pltrow.append(retval.variables[pred]['data'][(retval.variables['_sample_ind']['data'] == 0) & (retval.variables[str(idcol)]['data'] == id)]) # pop
#        idx += 1
#    ax = axs[row]
#    bplot = ax.boxplot(pltrow, patch_artist=True)
#    for patch, colour in zip(bplot['boxes'], ['pink', 'green']*ncols):
#        patch.set_facecolor(colour)
#
#canvas = FigureCanvasAgg(fig)
#fig.tight_layout()
#
#buf = BytesIO()
#canvas.print_figure(buf, dpi=80, format='svg')
#
#buf.seek(0)
#outputs['box_alt.svg'] = ImageOutput(buf.getvalue())
#buf.close()


#idx = 0
#for row in range(0, nrows):
#    fig = Figure(figsize=(8,8))
#    ax = fig.add_subplot(100 + 10 + 1)
#    pltrow = []
#    idxstart = idx
#    for col in range(0, ncols):
#        if idx >= len(uniqids):
#                continue
#        id = uniqids[idx]
#        #print("row: " + str(row) + " col: " + str(col) + " idx: " + str(idx) + " id: " + str(id))
#        pltrow.append(retval.variables[pred]['data'][(retval.variables['_sample_ind']['data'] == 1) & (retval.variables[str(idcol)]['data'] == id)]) # samp
#        pltrow.append(retval.variables[pred]['data'][(retval.variables['_sample_ind']['data'] == 0) & (retval.variables[str(idcol)]['data'] == id)]) # pop
#        idx += 1
#    bplot = ax.boxplot(pltrow, patch_artist=True)
#    for patch, colour in zip(bplot['boxes'], ['pink', 'green']*ncols):
#        patch.set_facecolor(colour)
#
#    ax.set_title(pred)
#    xvals = numpy.arange(0, ncols * 2, 2) + 1.5
#    #print('x-values: ' + str(xvals))
#    ax.set_xticks(xvals)
#    if lab in retval.valuelabels and id in retval.valuelabels[lab]:
#        xlabs = [retval.valuelabels[lab][id] for id in uniqids[idxstart:idx]]
#    else:
#        xlabs = [str(idcol) + '==' + str(id) for id in uniqids[idxstart:idx]]
#    #print('x-labels: ' + str(xlabs))
#    ax.set_xticklabels(xlabs, rotation=-30, ha='left') # rotation='vertical')
#    ax.set_xmargin(0.2)
#
#    canvas = FigureCanvasAgg(fig)
#    fig.tight_layout()
#
#    buf = BytesIO()
#    canvas.print_figure(buf, dpi=80, format='svg')
#
#    buf.seek(0)
#    outputs['box_' + str(row) + '.svg'] = ImageOutput(buf.getvalue())
#    buf.close()

if not (numpy.all(sampvars == sampvars[0]) or numpy.all(popvars == popvars[0])): # Ignore constant columns

    mpop = RunStatJR(template='2LevelMod', dataset='workdata2', invars = {'y': str(pred), 'L2ID': str(idcol), 'storeresid': 'No', 'D': 'Normal', 'x': 'Intercept'}, estoptions = {'Engine': 'R_lme4', 'defaultalg': 'Yes', 'nAGQ': '1'})
    varpop2 = mpop.eng.outputs['ModelResults'].content['paramstats']['est'][str(idcol) + '__Intercept_']
    varpop1 = mpop.eng.outputs['ModelResults'].content['paramstats']['est']['Residual']
    vpcpop = varpop2 / (varpop2 + varpop1)

    msamp = RunStatJR(template='2LevelMod', dataset='workdata', invars = {'y': str(pred), 'L2ID': str(idcol), 'storeresid': 'No', 'D': 'Normal', 'x': 'Intercept'}, estoptions = {'Engine': 'R_lme4', 'defaultalg': 'Yes', 'nAGQ': '1'})
    varsamp2 = msamp.eng.outputs['ModelResults'].content['paramstats']['est'][str(idcol) + '__Intercept_']
    varsamp1 = msamp.eng.outputs['ModelResults'].content['paramstats']['est']['Residual']
    vpcsamp = varsamp2 / (varsamp2 + varsamp1)

    html = '<p>'
    html += 'Here for the predictor variable ' + str(pred) + ' the VPC is ' + str(numpy.round(vpcpop, 2)) + ' in the population dataset and ' + str(numpy.round(vpcsamp, 2)) + ' in the sample dataset. '
    html += '</p>'

    outputs['vpctext.html'] = HTMLOutput(html, "VPC text for predictor")

html = '<p>'
html += 'We will next look at the data in the individual small areas, and in particular the mean and standard deviation of the variable ' + str(pred) + ' for both the sample and the population dataset. '
html += 'Here aside from some areas having no sample data we can look and see how representative the samples in other areas are of the population. '
html += '</p>'

outputs['summarytext.html'] = HTMLOutput(html, "Summary text for predictor")


tabout = TabularOutput()
tabout.column_headings = ['Code', 'Name', 'Sample mean', 'Sample sd', 'Population mean', 'Population sd']

for id in uniqids:
    if lab in retval.valuelabels and id in retval.valuelabels[lab]:
        name = retval.valuelabels[lab][id]
    else:
        name = str(idcol) + '==' + str(id)
    sampmean = retval.variables[pred]['data'][(retval.variables['_sample_ind']['data'] == 1) & (retval.variables[str(idcol)]['data'] == id)].mean()
    sampsd = retval.variables[pred]['data'][(retval.variables['_sample_ind']['data'] == 1) & (retval.variables[str(idcol)]['data'] == id)].std()
    popmean = retval.variables[pred]['data'][(retval.variables['_sample_ind']['data'] == 0) & (retval.variables[str(idcol)]['data'] == id)].mean()
    popsd = retval.variables[pred]['data'][(retval.variables['_sample_ind']['data'] == 0) & (retval.variables[str(idcol)]['data'] == id)].std()

    tabout.add_row(str(id), [name, sampmean, sampsd, popmean, popsd])

outputs['summary'] = tabout

html = '<p>'
html += 'We can also look at the distribution of the predictor ' + str(pred) +' for the sample and population in each district as shown in the following plot:'
html += '</p>'

outputs['boxtext.html'] = HTMLOutput(html, "Box text for predictor")

# TODO: Determine dimensions from number of records
fig = Figure(figsize=(8,24))
ax = fig.add_subplot(100 + 10 + 1)
pltrow = []
for id in uniqids:
    pltrow.append(retval.variables[pred]['data'][(retval.variables['_sample_ind']['data'] == 1) & (retval.variables[str(idcol)]['data'] == id)]) # samp
    pltrow.append(retval.variables[pred]['data'][(retval.variables['_sample_ind']['data'] == 0) & (retval.variables[str(idcol)]['data'] == id)]) # pop
bplot = ax.boxplot(pltrow, vert=False, patch_artist=True)
for patch, colour in zip(bplot['boxes'], ['pink', 'green']*len(uniqids)):
    patch.set_facecolor(colour)

ax.set_xlabel(pred)
yvals = numpy.arange(0, len(uniqids) * 2, 2) + 1.5
ax.set_yticks(yvals)
if lab in retval.valuelabels and id in uniqids:
    ylabs = [retval.valuelabels[lab][id] for id in uniqids]
else:
    ylabs = [str(idcol) + '==' + str(id) for id in uniqids]
ax.set_yticklabels(ylabs)

ax.legend(handles=[patches.Patch(color='pink', label='Sample'), patches.Patch(color='green', label='Population')], bbox_to_anchor=(0., 1.02, 1., .102), loc=3, ncol=2, mode="expand", borderaxespad=0.)

canvas = FigureCanvasAgg(fig)
fig.tight_layout()

buf = BytesIO()
canvas.print_figure(buf, dpi=80, format='svg')

buf.seek(0)
outputs['box_vert.svg'] = ImageOutput(buf.getvalue())
buf.close()
'''